/****************************************************************************

 ■PhotoFrame  ControlBoard Version 1.0
		Copyright(C) 2018 Denshi Usagi

****************************************************************************/

// PIC16F1827 Configuration Bit Settings

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config CPD = OFF        // Data Memory Code Protection (Data memory code protection is disabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable (Brown-out Reset disabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = ON        // Internal/External Switchover (Internal/External Switchover mode is enabled)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is enabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config PLLEN = OFF      // PLL Enable (4x PLL disabled)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LVP = ON         // Low-Voltage Programming Enable (Low-voltage programming enabled)

#include <xc.h>
#include "typedefs.h"

//=================================================================

#define RSP_VCC			LATA1
#define LCD_VCC			LATA2
#define	RSP_GPIO9		RA7
#define	RSP_GPIO10		RA0
#define	RSP_GPIO11		LATA6
#define	PSW				RA3
#define	POW				LATA4
#define LED				LATB1
#define LCD_KEY1		LATB2
#define LCD_KEY2		LATB3
#define	RTC_INT			RB0

#define AN_MIC			0b00111 // AN7
#define AN_CDS			0b01000 // AN8

#define RSP_O_STOP		RSP_GPIO11
#define RSP_I_EXIT		RSP_GPIO10
#define RSP_I_LCD		RSP_GPIO9

#define EEP_FLAG		0	// EEPROMアドレス(初回起動フラグ)
#define EEP_BLT			2	// EEPROMアドレス(現在のブライトネス設定)
#define EEP_CDS_THR		4	// EEPROMアドレス(CDS閾値設定値)
#define EEP_MIC_THR		6	// EEPROMアドレス(MIC閾値設定値)
#define EEP_CDS_MAX		8	// EEPROMアドレス(CDS値のMAX設定値))

#define MODE_STANDBY	0	// 待機状態
#define MODE_OPERATION	1	// 稼働状態
#define MODE_SETUP		2	// 設定状態
#define MODE_SHUTDOWN	3	// シャットダウンする

#define SW_NONE			0
#define SW_PUSH			1
#define SW_LONG			2

#define LCD_OFF			0
#define LCD_ON			1
#define LCD_STANDBY		2

#define CDS_DEF_THR		0x037	// デフォルトCds閾値(以下で消灯と判断)
#define MIC_DEF_THR		0x210	// デフォルトマイク閾値(以下で無音と判断)
#define CDS_DEF_MAX		0x2C0	// デフォルトCDS値のMAX値(十分に明るい時のs_wCDSの値)

#define CDS_CNT_THR_U	5		// 連続換算で5秒間 … 1s就寝検出がこの回数を超えたら待機モードへ
#define CDS_CNT_THR_L	1		// 1s就寝検出がこの回数を下回ったら復帰する
#define CDS_CNT_THR_M	60		// 最大カウント回数

#define MIC_DOWN		( 1 * 60)	// 音声を検出した時に減算する無音カウント値
#define MIC_CNT_THR_U	(15 * 60)	// 連続換算で15分間 … 1s無音検出がこの回数を超えたら待機モードへ
#define MIC_CNT_THR_L	(MIC_CNT_THR_U - MIC_DOWN * 5) // … 1s無音検出がこの回数を下回ったら復帰する

//=================================================================

static void InitSystem(void);
static void InitApp(void);
static void Run(void);

static int ModeOperation(void);
static BOOL Brightness(void);
static int ModeStandby(void);
static void ModeSetup(void);
static void SetupMicThr(void);
static void SetupCdsThr(void);
static void SetupCdsMax(void);
static void SetupBlt(void);
static int GetSwitch(void);
static void BlinkLED(BYTE bS, BYTE bT1, BYTE bT2);
static void RaspiCtrl(int nCtrl);
static void PanelCtrl(int nCtrl);
static BOOL BackLight(int nCtrl);
static void Shutdown(void);

static WORD GetCDS(void);
static WORD GetMIC(void);
static void CheckMIC(void);
static WORD GetElapsed(WORD wBase);
static void WriteEEP8(int nAdd, BYTE nData);
static void WriteEEP16(int nAdd, WORD nData);
static BYTE ReadEEP8(int nAdd);
static WORD ReadEEP16(int nAdd);
static void WaitMs(WORD wMs);
static void WaitSec(WORD wSec);

//=================================================================

static WORD s_wTick = 0;
static WORD s_wWait;
static BYTE s_LedFR;
static BYTE s_LedCN;
static BYTE s_LedT1;
static BYTE s_LedT2;
static BYTE s_LedST;
static BYTE s_LedWK;

static int s_nLcd;
static int s_nBlt;
static int s_nMode;

static WORD s_wThrCds;
static WORD s_wThrMic;
static WORD s_wMaxCds;
static WORD s_wCntCds;
static WORD s_wCntMic;
static WORD s_wCDS;
static WORD s_wAvgCDS;
static WORD s_wMIC;
static BOOL s_fMIC;

//=================================================================

// ●メインエントリ
void main(void)
{
	InitSystem();
	InitApp();
	Run();
}

// ●システム初期化
static void InitSystem(void)
{
	// クロック設定
	OSCCONbits.IRCF = 0b0111;// 500 kHz MF

	// ポート初期化
	PORTA = PORTB = 0x00;
	LATA = LATB = 0x00;

	// POW
	POW = 1; // Powerup latch
	ANSELAbits.ANSA4 = 0;
	TRISAbits.TRISA4 = 0;

	// RSP_VCC
	ANSELAbits.ANSA1 = 0;
	TRISAbits.TRISA1 = 0;

	// LCD_VCC
	ANSELAbits.ANSA2 = 0;
	TRISAbits.TRISA2 = 0;
	
	// RSP_GPIO9
	TRISAbits.TRISA7 = 1;

	// RSP_GPIO10
	ANSELAbits.ANSA0 = 0;
	TRISAbits.TRISA0 = 1;

	// RSP_GPIO11
	TRISAbits.TRISA6 = 0;

	// PSW
	ANSELAbits.ANSA3 = 0;
	TRISAbits.TRISA3 = 1;

	// LED
	LED = 1; // LED ON
	ANSELBbits.ANSB1 = 0;
	TRISBbits.TRISB1 = 0;

	// LCD_KEY1
	ANSELBbits.ANSB2 = 0;
	TRISBbits.TRISB2 = 0;

	// LCD_KEY2
	ANSELBbits.ANSB3 = 0;
	TRISBbits.TRISB3 = 0;
	
	// RTC_INT
	TRISBbits.TRISB0 = 1;

	// MIC/CDS
	ANSELBbits.ANSB4 = 1;
	ANSELBbits.ANSB5 = 1;

	// ADC
	ADCON1bits.ADFM = 1;
	ADCON1bits.ADCS = 0b111; // FRC
	ADCON1bits.ADPREF = 0b00; // AVDD
	ADCON0bits.ADON = 1;

	// Timer1 (LEDブリンク用)
	T1CONbits.TMR1CS = 0b01; // FOSC
	T1CONbits.T1CKPS = 0b00; // 1:1

	// Timer2 (TickCount用)
	T2CONbits.T2CKPS = 0b00; // 1:1
	T2CONbits.T2OUTPS = 0b0000; // 1:1
	PR2 = 125; // 1ms
	T2CONbits.TMR2ON = 1;
	TMR2IE = 1;
	
	// Timer4 (WaitMs用)
	T4CONbits.T4CKPS = 0b00; // 1:1
	T4CONbits.T4OUTPS = 0b0000; // 1:1
	PR4 = 125; // 1ms

	// Timer6 (監視用)
	T6CONbits.T6CKPS = 0b11; // 1:64
	T6CONbits.T6OUTPS = 0b1000; // 1:9
	PR6 = 217; // ≒1000ms

	// 割込み許可
	//INTE = 1; // 当初使おうと考えたがやめた
	PEIE = 1;
	GIE = 1;
}

// ●アプリ初期化
static void InitApp(void)
{
#if 0 // 火入れ時にデバッガで確認
	RSP_VCC = 1;
	LCD_VCC = 1;

	RSP_GPIO11 = 1;
	RSP_GPIO11 = 0;
	LCD_KEY1 = 1;
	LCD_KEY1 = 0;
	LCD_KEY2 = 1;
	LCD_KEY2 = 0;

	BYTE aaa = RSP_GPIO9;
	BYTE bbb = RSP_GPIO10;
	BYTE ccc = PSW;

	for(;;)
	{
		WORD cds = GetCDS();
		WORD mic = GetMIC();
		
		//LED = !LED; WaitMs(1000); //Lチカ
		LED = (mic >= MIC_DEF_THR);
	}
#endif

	// SWを離すまで待つ
	while (PSW);
	WaitSec(1);
	
	s_nLcd = LCD_OFF;

	if (ReadEEP16(EEP_FLAG) != 77)
	{
		// 初起動 EEPROM初期化
		s_nBlt = 0;
		s_wThrCds = CDS_DEF_THR;
		s_wThrMic = MIC_DEF_THR;
		s_wMaxCds = CDS_DEF_MAX;
		
		WriteEEP16(EEP_BLT, (BYTE)s_nBlt);
		WriteEEP16(EEP_CDS_THR, s_wThrCds);
		WriteEEP16(EEP_MIC_THR, s_wThrMic);
		WriteEEP16(EEP_CDS_MAX, s_wMaxCds);
		WriteEEP16(EEP_FLAG, 77);
	}
	else
	{
		s_nBlt = ReadEEP16(EEP_BLT);
		s_wThrCds = ReadEEP16(EEP_CDS_THR);
		s_wThrMic = ReadEEP16(EEP_MIC_THR);
		s_wMaxCds = ReadEEP16(EEP_CDS_MAX);
	}

	// MICとCDSの監視を開始
	s_wCntCds = 0;
	s_wCntMic = 0;
	s_wCDS = 0;
	s_wAvgCDS = 0;
	s_wMIC = 0;
	s_fMIC = FALSE;

	TMR6IE = 1;
	TMR6ON = 1;	
}

// ●実行ループ
static void Run(void)
{
	// まずは稼働モードで起動
	for(int nNext = MODE_OPERATION;;)
	{
		BlinkLED(3, 1, 0);

		switch (nNext)
		{
		case MODE_OPERATION:
			s_nMode = nNext;
			nNext = ModeOperation();
			break;

		case MODE_STANDBY:
			s_nMode = nNext;
			nNext = ModeStandby();
			break;
		
		case MODE_SETUP:
			ModeSetup();
			nNext = s_nMode;
			break;
		
		case MODE_SHUTDOWN:
			Shutdown();
			break;
		}
	}
}

//------------------------------------------------------------------------------

// ●稼働モード
static int ModeOperation(void)
{
	// ラズパイ電源ON
	RaspiCtrl(1);
	
	BlinkLED(0, 0, 0);
	for(;;)
	{
		// スイッチ入力
		switch (GetSwitch())
		{
		// プッシュで設定モード
		case SW_PUSH:
			if (!RSP_O_STOP) {
				return MODE_SETUP;
			}
			break;
			
		// 長押しでシャットダウン
		case SW_LONG:
			return MODE_SHUTDOWN;
		}

		// ラズパイからのパネル制御
		PanelCtrl((RSP_I_LCD)? LCD_ON : LCD_STANDBY);

		// ラズパイからの終了通知
		if (RSP_I_EXIT) {
			return MODE_STANDBY;
		}

		// 稼働条件チェック(ラズパイへの終了指示)
		if (!RSP_O_STOP)
		{
			if (s_wCntCds >= CDS_CNT_THR_U || s_wCntMic >= MIC_CNT_THR_U)
			{
				RSP_O_STOP = 1;
				BlinkLED(3, 1, 0);
			}
		}

		// マイクチェック
		CheckMIC();

		if (!RSP_O_STOP)
		{
			if (s_nLcd == LCD_ON)
				Brightness();

			// 検知音をLEDに反映
			LED = s_fMIC;
		}
	}
}

// ●バックライト明るさ調整
static BOOL Brightness(void)
{
	WORD wLV = 0;
	if (s_wAvgCDS > s_wThrCds) {
		wLV = s_wAvgCDS - s_wThrCds;
	}
	
	#define BLV5	10	// 明るさ5の範囲
	#define BLV4	10	// 明るさ4の範囲
	#define BLV3	10	// 明るさ3の範囲
	#define BLV2	7	// 明るさ2の範囲(一番変化が大きい)
	#define BLV1	1	// 明るさ1の範囲(かなり暗いのでスレッショルド付近でこれにする)
	#define BL_MAX	(BLV1 + BLV2 + BLV3 + BLV4 + BLV5)
	#define CDS_VAL	((s_wMaxCds > s_wThrCds)? (s_wMaxCds - s_wThrCds) : 0)

	#define CTRL1	(BLV1 * CDS_VAL / BL_MAX)
	#define CTRL2	((BLV1 + BLV2) * CDS_VAL / BL_MAX)
	#define CTRL3	((BLV1 + BLV2 + BLV3) * CDS_VAL / BL_MAX)
	#define CTRL4	((BLV1 + BLV2 + BLV3 + BLV4) * CDS_VAL / BL_MAX)
	
	if (wLV <= CTRL1) return BackLight(1);
	if (wLV <= CTRL2) return BackLight(2);
	if (wLV <= CTRL3) return BackLight(3);
	if (wLV <= CTRL4) return BackLight(4);
	return BackLight(5);
}

// ●待機モード
static int ModeStandby(void)
{
	// パネルスタンバイ
	PanelCtrl(LCD_STANDBY);
	
	// ラズパイ電源OFF
	RaspiCtrl(0);
	
	BlinkLED(1, 1, 20);
	for(;;)
	{
		// スイッチ入力
		switch (GetSwitch())
		{
		// プッシュで設定モード
		case SW_PUSH:
			return MODE_SETUP;
			
		// 長押しでシャットダウン
		case SW_LONG:
			return MODE_SHUTDOWN;
		}
		
		if (s_wCntCds < CDS_CNT_THR_L && s_wCntMic < MIC_CNT_THR_L)
		{
			// 稼働条件復帰で起きる
			if (RSP_O_STOP) {
				return MODE_OPERATION;
			}
			// RTCアラームで起きる
			if (!RTC_INT) {
				return MODE_OPERATION;
			}
		}

		// 就寝に入ったらパネルの電源を切る
		if (s_nLcd == LCD_STANDBY)
		{
			if (s_wCntCds >= CDS_CNT_THR_U) {
				PanelCtrl(LCD_OFF);
			}
		}

		// マイクチェック
		CheckMIC();
	}
}

// ●設定モード
static void ModeSetup(void)
{
	for(;;)
	{
		// 連続点滅(戻る)
		BlinkLED(1, 1, 0);
		for(;;)
		{
			switch (GetSwitch()) {
			default: continue;
			case SW_PUSH: break;
			case SW_LONG: return;
			}
			break;
		}

		SetupMicThr();	// 1回点滅
		SetupCdsThr();	// 2回点滅
		SetupCdsMax();	// 3回点滅
		
		// バックライトの設定はONのときのみ
		if (s_nLcd == LCD_ON) {
			SetupBlt();	// 4回点滅
		}
	}
}

// ●マイク閾値調整操作
static void SetupMicThr(void)
{
	BlinkLED(1, 1, 7);

	for(;;)
	{
		switch (GetSwitch()) {
		case SW_PUSH: return;
		case SW_LONG:
			// 設定実行
			BlinkLED(6, 10, 10);

			WaitSec(11);
			BlinkLED(0, 0, 0);
			LED = 1;
			
			// 検出期間内での最大値を取得して保存する
			WORD wMax = 0;
			for(int i = 0; i < 80; i++) // 約4秒間 40サンプル
			{
				WaitMs(50);
				if (s_wMIC > wMax)
					wMax = s_wMIC;
			}
			s_wThrMic = wMax;
			WriteEEP16(EEP_MIC_THR, s_wThrMic);

			LED = 0;
			WaitSec(1);
			BlinkLED(1, 1, 7);
			continue;
		}
	}
}

// ●光量閾値調整操作
static void SetupCdsThr(void)
{
	BlinkLED(1, 2, 7);
	
	for(;;)
	{
		switch (GetSwitch()) {
		case SW_PUSH: return;
		case SW_LONG:
			// 設定実行
			BlinkLED(6, 2, 10);

			WaitSec(3);
			BlinkLED(0, 0, 0);
			LED = 1;
			
			// 検出期間内での最大値を取得して保存する
			WORD wMax = 0;
			for(int i = 0; i < 8; i++) // 約4秒間 40サンプル
			{
				WaitMs(500);
				if (s_wCDS > wMax)
					wMax = s_wCDS;
			}
			s_wThrCds = wMax;
			WriteEEP16(EEP_CDS_THR, s_wThrCds);
			
			LED = 0;
			WaitSec(1);
			BlinkLED(1, 2, 7);
			continue;
		}
	}
}

// ●最大光量調整操作
static void SetupCdsMax(void)
{
	BlinkLED(1, 3, 7);
	
	for(;;)
	{
		switch (GetSwitch()) {
		case SW_PUSH: return;
		case SW_LONG:
			// 設定実行
			BlinkLED(6, 2, 10);

			WaitSec(3);
			BlinkLED(0, 0, 0);
			LED = 1;
			
			// 検出期間内での最大値を取得して保存する
			WORD wMax = 0;
			for(int i = 0; i < 8; i++) // 約4秒間 4サンプル
			{
				WaitMs(500);
				if (s_wCDS > wMax)
					wMax = s_wCDS;
			}
			s_wMaxCds = wMax;
			WriteEEP16(EEP_CDS_MAX, s_wMaxCds);

			LED = 0;
			WaitSec(1);
			BlinkLED(1, 3, 7);
			continue;
		}
	}
}

// ●バックライト原点設定操作
static void SetupBlt(void)
{
	BlinkLED(1, 4, 6);

	for(;;)
	{
		switch (GetSwitch()) {
		case SW_PUSH: return;
		case SW_LONG:
			// 設定実行
			BlinkLED(6, 2, 10);

			WaitSec(3);
			BlinkLED(0, 0, 0);
			BackLight(5);
			LED = 1;
			
			// SWを押して最大の明るさにする
			for(;;)
			{
				switch (GetSwitch()) {
				case SW_PUSH:
					LCD_KEY2 = 1; // KEY2立ち下がりで切り替わり
					WaitMs(200);
					LCD_KEY2 = 0;
					WaitMs(200);					
				default:
					continue;
				case SW_LONG:
					break;
				}
				break;
			}
			s_nBlt = 0;
			WriteEEP16(EEP_BLT, (BYTE)s_nBlt);

			LED = 0;
			WaitSec(1);
			BlinkLED(1, 4, 6);
			continue;
		}
	}
}

// ●シャットダウン
static void Shutdown(void)
{
	// パネルの電源を切る
	PanelCtrl(LCD_OFF);

	if (RSP_VCC)
	{
		// ラズパイへの終了指示
		RSP_O_STOP = 1;

		// ラズパイからの終了通知待ち(タイムアウト30秒)
		WORD wBase = s_wTick;
		while (!RSP_I_EXIT && GetElapsed(wBase) < (30 * 1000));

		// ラズパイ電源OFF
		RaspiCtrl(0);
	}
	
	// 終了処理
	BlinkLED(0, 0, 0);
	GIE = 0;

	// 電源OFF
	POW = 0;
	
	// 電断待ち
	for(;;) SLEEP();
}

//------------------------------------------------------------------------------

// ●スイッチ入力
static int GetSwitch(void)
{
	for(int i = 0; i < 500; i++)
	{
		if (PSW)
		{
			WORD wBase = s_wTick;
			while(PSW);
			
			// 600ms以上押すと長押し
			return (GetElapsed(wBase) < 600)? SW_PUSH : SW_LONG;
		}
		// マイクチェック
		CheckMIC();
	}
	return SW_NONE;
}

// ●LEDブリンク(bS:1サイクル長 bT1:点滅回数 bT2:消灯期間)
static void BlinkLED(BYTE bS, BYTE bT1, BYTE bT2)
{
	TMR1ON = 0;	
	TMR1IE = 0;
	TMR1IF = 0;
	LED = 0;

	if (bS == 0 || bT1 == 0) {
		return;
	}

	s_LedFR = bS;
	s_LedCN = 0;
	s_LedT1 = bT1;
	s_LedT2 = bT2;
	s_LedST = 0;
	s_LedWK = 0;

	LED = 1;
	TMR1 = 0;
	TMR1IE = 1;
	TMR1ON = 1;	
}

// ●ラズパイの電源制御(0:OFF, 1:ON)
static void RaspiCtrl(int nCtrl)
{
	if (RSP_VCC != nCtrl)
	{
		if (nCtrl == 1)
		{
			RSP_O_STOP = 0;
			RSP_VCC = 1;
			WaitSec(30);// 起動に30秒
		}
		else
		{
			WaitSec(15); // シャットダウンに15秒
			RSP_VCC = 0;
			WaitMs(500);
		}
	}
}

// ●液晶パネルの制御(0:OFF, 1:ON, 2:STANDBY)
static void PanelCtrl(int nCtrl)
{
	if (s_nLcd != nCtrl)
	{
		switch (nCtrl)
		{
		case LCD_ON:
			if (s_nLcd == LCD_OFF)
			{
				// OFF → ON
				LCD_VCC = 1;
				WaitSec(2); // ラズパイを先に起動してある場合、表示までに2秒
			}
			else
			{
				// STANDBY → ON
				LCD_KEY1 = 1;
				WaitMs(300);
				LCD_KEY1 = 0; // KEY1立ち上がりで復帰
				WaitSec(2);
			}
			s_nLcd = nCtrl;
			break;

		case LCD_STANDBY:
			// ON → STANDBY
			if (s_nLcd == LCD_ON)
			{
				LCD_KEY1 = 1; // KEY1立ち下がりでスタンバイ
				WaitMs(500);
				LCD_KEY1 = 0;
				WaitSec(1);
				s_nLcd = nCtrl;
			}
			break;

		case LCD_OFF:
			if (s_nLcd == LCD_ON)
			{
				// ON → STANDBY
				LCD_KEY1 = 1; // KEY1立ち下がりでスタンバイ
				WaitMs(500);
				LCD_KEY1 = 0;
				WaitSec(1);
			}
			// STANDBY → OFF
			LCD_VCC = 0;
			WaitSec(3);  // OFFしてすぐにONするとパネルがバグる
			s_nLcd = nCtrl;
			break;
		}
	}
}

// ●バックライト明るさ(1～5)
static BOOL BackLight(int nCtrl)
{
    static int sTbl[8] = { 5, 4, 3, 2, 1, 2, 3, 4 };
	
	if (s_nLcd == LCD_ON)
	{
		if (sTbl[(unsigned)s_nBlt] != nCtrl)
		{
			do {
				s_nBlt = (s_nBlt + 1) % 8;

				LCD_KEY2 = 1; // KEY2立ち下がりで切り替わり
				WaitMs(200);
				LCD_KEY2 = 0;
				WaitMs(400);
			}
			while (sTbl[(unsigned)s_nBlt] != nCtrl);

			WriteEEP16(EEP_BLT, (unsigned)s_nBlt);

			return TRUE;
		}
	}
	
	return FALSE;
}

// ●Cds電圧を取得
static WORD GetCDS(void)
{
	ADCON0bits.CHS = AN_CDS;

	ADCON0bits.GO_nDONE = 1;
	while (ADCON0bits.GO_nDONE);

	return (WORD)ADRESH << 8 | ADRESL;
}

// ●マイク電圧を取得
static WORD GetMIC(void)
{
	ADCON0bits.CHS = AN_MIC;

	ADCON0bits.GO_nDONE = 1;
	while (ADCON0bits.GO_nDONE);

	return (WORD)ADRESH << 8 | ADRESL;
}

// ●マイク電圧をチェック
static void CheckMIC(void)
{
	s_wMIC = GetMIC();
	if (s_wMIC > s_wThrMic)
		s_fMIC = TRUE;
}

// ●経過時間を取得する(1ms単位)
static WORD GetElapsed(WORD wBase)
{
	WORD wTick = s_wTick;
	if (wTick >= wBase) {
		return (wTick - wBase);
	}

	WORD wRet = 65535U - (wBase - wTick);
	return wRet + 1;
}

// ●EEPROMへ書き出し
static void WriteEEP8(int nAdd, BYTE nData)
{
    BYTE GIEBitValue = 0;

    EEADRL = (BYTE)nAdd;
    EEDATL = nData;
    EECON1bits.EEPGD = 0;
    EECON1bits.CFGS = 0;
    EECON1bits.WREN = 1;

    GIEBitValue = INTCONbits.GIE;
    INTCONbits.GIE = 0;
    EECON2 = 0x55;
    EECON2 = 0xAA;
    EECON1bits.WR = 1;

    while (EECON1bits.WR);

    EECON1bits.WREN = 0;
    INTCONbits.GIE = GIEBitValue;
}

// ●EEPROMへ書き出し
static void WriteEEP16(int nAdd, WORD nData)
{
	WriteEEP8(nAdd + 0, (BYTE)nData);
	WriteEEP8(nAdd + 1, (BYTE)(nData >> 8));
}

// ●EEPROMから読み出し
static BYTE ReadEEP8(int nAdd)
{
    EEADRL = (BYTE)nAdd;
    EECON1bits.CFGS = 0;
    EECON1bits.EEPGD = 0;
    EECON1bits.RD = 1;
    NOP();
    NOP();

    return (EEDATL);
}

// ●EEPROMから読み出し
static WORD ReadEEP16(int nAdd)
{
	WORD wRet = ReadEEP8(nAdd + 0);
	wRet |= (unsigned)(ReadEEP8(nAdd + 1) << 8);
	return wRet;
}

// ●ミリ秒待ち(max. 65535ms))
static void WaitMs(WORD wMs)
{
	s_wWait = wMs;
	TMR4 = 0;
	TMR4IE = 1;
	TMR4ON = 1;
	while(TMR4IE);
}

// ●秒待ち(max. 65535s))
static void WaitSec(WORD wSec)
{
	do WaitMs(1000);
	while (--wSec > 0);
}

// ●割込み
void interrupt Interrupt(void)
{
	if (TMR1IF) // LEDブリンク ≒130ms
	{
		if (++s_LedCN >= s_LedFR)
		{
			if (s_LedST == 0)
			{
				// 点滅期間
				if (LED == 1)
				{
					LED = 0;
				}
				else
				{
					if (++s_LedWK >= s_LedT1)
					{
						if (s_LedT2 == 0)
						{
							LED = 1;
						}
						else
						{
							s_LedST = 1;
						}
						s_LedWK = 0;
					}
					else
					{
						LED = 1;
					}
				}
			}
			else
			{
				// 休止期間
				if (++s_LedWK >= s_LedT2)
				{
					LED = 1;
					s_LedST = 0;
					s_LedWK = 0;
				}
			}
			s_LedCN = 0;
		}
		TMR1IF = 0;
	}

	else if (TMR2IF) // 1ms単位
	{
		s_wTick++;
		TMR2IF = 0;
	}

	else if (TMR4IF) // 1ms単位
	{
		if (--s_wWait == 0)
		{
			TMR4ON = 0;
			TMR4IE = 0;
		}
		TMR4IF = 0;
	}

	else if (TMR6IF) // 1000ms単位
	{
		// CDS
		s_wCDS = (s_wCDS + GetCDS()) / 2;
		if (s_wCDS <= s_wThrCds) {
			if (s_wCntCds < CDS_CNT_THR_M) ++s_wCntCds;
		}
		else {
			if (s_wCntCds > 0) --s_wCntCds;
		}

		// AvgCDS(1分間の平均)
		static WORD s_wVal[60] = { 0 };
		static BYTE s_nIndex = -1;
		static BYTE s_nCount = 0;
		
		if (++s_nIndex > 59) s_nIndex = 0;
		if (++s_nCount > 60) s_nCount = 60;

		s_wVal[s_nIndex] = s_wCDS;

		DWORD dwTotal = 0;
		for(BYTE i = 0; i < s_nCount; i++)
			dwTotal += s_wVal[i];

		s_wAvgCDS = (WORD)(dwTotal / s_nCount);

		// MIC
		if (!s_fMIC) {
			if (s_wCntMic < MIC_CNT_THR_U) ++s_wCntMic;
		}
		else
		{
			s_fMIC = FALSE;
			if (s_wCntMic <= MIC_DOWN) s_wCntMic = 0;
			else s_wCntMic -= MIC_DOWN;
		}
		
		TMR6IF = 0;
	}
}
