/****************************************************************************

 ■PhotoFrame
	アプリケーション
        Copyright(C) 2018 Denshi Usagi

****************************************************************************/

#include "main.h"
#include "Application.h"
#include "SDL2/SDL.h"

// 実行状態
enum EAppState
{
	// SC_PotoDisplay
	ST_DisplayPhoto,
	ST_DisplayPhotoWait,
	ST_DisplayCalender,
	ST_DisplayCalenderWait,
	ST_GetWeatherInfo,
	ST_DisplayWeather,
	ST_DisplayWeatherWait,
	ST_GetKosaInfo,
	ST_DisplayKosa,
	ST_DisplayKosaWait,

	// SC_SpecialPhoto
	ST_DispSpecialPhoto,
	ST_DispSpecialPhotoWait,
};

//---------------------------------------------------------------------------

// ●コンストラクタ
CApplication::CApplication()
{
	m_nStatePhoto = ST_DisplayCalender;
	m_nStateSpecial = ST_DispSpecialPhoto;
	m_nNextTime = 0;
	m_bDisplayed = false;
	m_bStartup = true;
	m_pMusic = NULL;
}

// ●初期化
bool CApplication::Initialize()
{
	if (!m_Display.Create())
		return false;

	if (!m_Scaduler.Initialize())
		return false;

	if (!m_PhotoList.Initialize())
		return false;

	m_Weather.Initialize();
	m_Kosa.Initialize();

	return true;
}

// ●終了
void CApplication::Finalize(void)
{
	if (m_pMusic != NULL)
	{
		// 再生中の場合は終わるまで待つ
		while (Mix_PlayingMusic())
			SDL_Delay(200);

		Mix_FreeMusic(m_pMusic);
		m_pMusic = NULL;
	}

	m_Display.Destroy();
}

// ●フォトフレーム処理
int CApplication::Process(void)
{
	// メッセージ処理
	if (PumpMessage())
		return 1; // ユーザー操作による終了

	// スケジュール(次回の起動アラームもセット)
	String sInfo;
	EScadule scd = m_Scaduler.Scadule(&sInfo);

	// 制御基板からの終了指示をチェック
	if (g_CtrlBoard.IsStop())
	{
		if (m_bDisplayed)
			m_Display.Ending();

		g_CtrlBoard.EnablePanel(false);
		return 2; 
	}

	switch (scd)
	{
	case SC_PotoDisplay:	// 通常動作
		g_CtrlBoard.EnablePanel(true);
		StartupSound();
		PhotoDisplay();
		m_bDisplayed = true;
		break;

	case SC_SpecialPhoto:	// 特別写真
		g_CtrlBoard.EnablePanel(true);
		StartupSound();
		SpecialPhoto(&sInfo);
		m_bDisplayed = true;
		break;

	case SC_PlaySound:		// サウンド再生
		g_CtrlBoard.EnablePanel(true);
		StartupSound();
		PlaySound(&sInfo);
		break;

	case SC_Wait:	// 何もせずに待つ
		StartupSound();
		if (m_bDisplayed)
		{
			m_Display.Ending();
			m_bDisplayed = false;
		}
		g_CtrlBoard.EnablePanel(false);
		break;

	default:	// 通常終了(スケジュールなし)
		if (m_bDisplayed)
			m_Display.Ending();

		g_CtrlBoard.EnablePanel(false);
		return 2; 
	}

	return 0;
}

// ●メッセージポンプ
bool CApplication::PumpMessage(void)
{
	for(;;)
    {
		// イベント取得
		SDL_Event ev;
		if (SDL_PollEvent(&ev) == 0)
		    break;
	
		// 終了操作をチェック
		switch(ev.type)
		{
		case SDL_MOUSEBUTTONUP:
			if (ev.button.button == SDL_BUTTON_RIGHT)
				return true;
			break;

		case SDL_QUIT:
			return true;
		}
    }

	return false;
}

// ●通常動作
void CApplication::PhotoDisplay(void)
{
	PCSTR pPhotoFile;

START:
	switch(m_nStatePhoto)
	{
	case ST_DisplayPhoto: // 写真表示
		pPhotoFile = m_PhotoList.GetNext();
		if (pPhotoFile != NULL)
		{
			m_Display.DisplayImage(pPhotoFile);
			m_nNextTime = time(NULL) + g_Settings.ReadInt("DisplayPhotoTime", 20);
			m_nStatePhoto++;
		}
		else
		{
			m_nStatePhoto = ST_DisplayCalender;
			goto START;
		}
		break;

	case ST_DisplayPhotoWait:
		if (time(NULL) >= m_nNextTime)
		{
			m_nStatePhoto++;
			goto START;
		}
		break;

	case ST_DisplayCalender: // カレンダー表示
		m_Display.DisplayCalender(m_Scaduler.GetTempture());
		m_nNextTime = time(NULL) + g_Settings.ReadInt("DisplayCalenderTime", 8);
		m_nStatePhoto++;
		break;

	case ST_DisplayCalenderWait:
		if (time(NULL) >= m_nNextTime)
		{
			m_nStatePhoto++;
			goto START;
		}
		break;

	case ST_GetWeatherInfo: // 天気予報取得
		m_Weather.Update();
		m_nStatePhoto = (m_Weather.IsEnable())? ST_DisplayWeather : ST_GetKosaInfo;
		goto START;

	case ST_DisplayWeather: // 天気予報表示
		m_Display.DisplayWeather(m_Weather.GetWeather());
		m_nNextTime = time(NULL) + g_Settings.ReadInt("DisplayWeatherTime", 10);
		m_nStatePhoto++;
		break;

	case ST_DisplayWeatherWait:
		if (time(NULL) >= m_nNextTime)
		{
			m_nStatePhoto++;
			goto START;
		}
		break;

	case ST_GetKosaInfo: // 黄砂予報取得
		m_Kosa.Update();
		m_nStatePhoto = (m_Kosa.IsEnable())? ST_DisplayKosa : ST_DisplayPhoto;
		goto START;

	case ST_DisplayKosa: // 黄砂予報表示
		m_Display.DisplayKosa(m_Kosa.GetKosa());
		m_nNextTime = time(NULL) + g_Settings.ReadInt("DisplayKosaTime", 10);
		m_nStatePhoto++;
		break;

	case ST_DisplayKosaWait:
		if (time(NULL) >= m_nNextTime)
		{
			m_nStatePhoto = ST_DisplayPhoto;
			goto START;
		}
		break;
	}
}

// ●特別動作(特別写真)
void CApplication::SpecialPhoto(const String* pInfo)
{
	switch(m_nStateSpecial)
	{
	case ST_DispSpecialPhoto:
		m_Display.DisplayImage(pInfo->c_str());
		m_nNextTime = time(NULL) + g_Settings.ReadInt("DisplaySpecialTime", 60);
		m_nStateSpecial++;
		break;

	case ST_DispSpecialPhotoWait:
		if (time(NULL) >= m_nNextTime)
		{
			m_Scaduler.Complete();
			m_nStateSpecial = ST_DispSpecialPhoto;
		}
		break;
	}
}

// ●スタートアップ時のサウンド再生
void CApplication::StartupSound(void)
{
	if (m_bStartup)
	{
		String path = g_Settings.ReadStr("StartupSound", "");
		if (path != "")
		{
			g_CtrlBoard.EnablePanel(true);
			PlaySound(&path);
			while (Mix_PlayingMusic())
				SDL_Delay(200);
		}
		m_bStartup = false;
	}
}

// ●特別動作(ウンド再生)
void CApplication::PlaySound(const String* pInfo)
{
	// 再生中の音声を停止
	if (m_pMusic != NULL)
	{
		Mix_FreeMusic(m_pMusic);
		m_pMusic = NULL;
	}

	// 音声ファイル(MP3)ロード
	m_pMusic = Mix_LoadMUS(pInfo->c_str());
	if (m_pMusic == NULL)
	{
		MessageBox("CApplication", Mix_GetError());
		return;
	}

	// 再生(非同期)
	if (Mix_PlayMusic(m_pMusic, 1) != 0)
	{
		MessageBox("CApplication", Mix_GetError());

		Mix_FreeMusic(m_pMusic);
		m_pMusic = NULL;
	}
}
