/****************************************************************************

 ■PhotoFrame
	RTC(RX8900)
        Copyright(C) 2018 Denshi Usagi

****************************************************************************/

#include "RTC.h"
#include "main.h"
#include <wiringPiI2C.h>
#include <stdio.h>

// RX-8900 I2Cスレーブアドレス
#define SLAVE_ADDR	0x32

// RX-8900 互換レジスタ
#define RX8900_BTC_SEC					0x00
#define RX8900_BTC_MIN					0x01
#define RX8900_BTC_HOUR					0x02
#define RX8900_BTC_WEEK					0x03
#define RX8900_BTC_DAY					0x04
#define RX8900_BTC_MONTH				0x05
#define RX8900_BTC_YEAR					0x06
#define RX8900_BTC_RAM					0x07
#define RX8900_BTC_ALARM_MIN			0x08
#define RX8900_BTC_ALARM_HOUR			0x09
#define RX8900_BTC_ALARM_WEEK_OR_DAY	0x0A
#define RX8900_BTC_TIMER_CNT_0			0x0B
#define RX8900_BTC_TIMER_CNT_1			0x0C
#define RX8900_BTC_EXT					0x0D
#define RX8900_BTC_FLAG					0x0E
#define RX8900_BTC_CTRL					0x0F

// RX-8900 拡張レジスタ
#define RX8900_EXT_SEC					0x10
#define RX8900_EXT_MIN					0x11
#define RX8900_EXT_HOUR					0x12
#define RX8900_EXT_WEEK					0x13
#define RX8900_EXT_DAY					0x14
#define RX8900_EXT_MONTH				0x15
#define RX8900_EXT_YEAR					0x16
#define RX8900_EXT_TEMP					0x17
#define RX8900_EXT_BACKUP				0x18
#define RX8900_EXT_TIMER_CNT_0			0x1B
#define RX8900_EXT_TIMER_CNT_1			0x1C
#define RX8900_EXT_EXT					0x1D
#define RX8900_EXT_FLAG					0x1E
#define RX8900_EXT_CTRL					0x1F

// RX8900_BTC_EXT レジスタビット位置
#define RX8900_BTC_EXT_TSEL0		(1 << 0)
#define RX8900_BTC_EXT_TSEL1		(1 << 1)
#define RX8900_BTC_EXT_FSEL0		(1 << 2)
#define RX8900_BTC_EXT_FSEL1		(1 << 3) 
#define RX8900_BTC_EXT_TE 			(1 << 4)
#define RX8900_BTC_EXT_USEL			(1 << 5) 
#define RX8900_BTC_EXT_WADA			(1 << 6)
#define RX8900_BTC_EXT_TEST			(1 << 7)

// RX8900_BTC_FLAG レジスタビット位置
#define RX8900_BTC_FLAG_VDET 		(1 << 0)
#define RX8900_BTC_FLAG_VLF 		(1 << 1)
#define RX8900_BTC_FLAG_AF 			(1 << 3)
#define RX8900_BTC_FLAG_TF 			(1 << 4)
#define RX8900_BTC_FLAG_UF 			(1 << 5)

// RX8900_BTC_CTRL レジスタビット位置
#define RX8900_BTC_CTRL_RESET 		(1 << 0)
#define RX8900_BTC_CTRL_AIE 		(1 << 3)
#define RX8900_BTC_CTRL_TIE 		(1 << 4)
#define RX8900_BTC_CTRL_UIE 		(1 << 5)
#define RX8900_BTC_CTRL_CSEL0 		(1 << 6)
#define RX8900_BTC_CTRL_CSEL1		(1 << 7)

// BCD変換
static BYTE bcd2bin (BYTE val) { return val - 6 * (val >> 4); }
static BYTE bin2bcd (BYTE val) { return val + 6 * (val / 10); }

//---------------------------------------------------------------------------

// ●コンストラクタ
CRTC::CRTC()
{
	m_tTimer = 0;
	m_I2C = -1;
}

// ●初期化
bool CRTC::Initialize()
{
	m_I2C = wiringPiI2CSetup(SLAVE_ADDR);
	if (m_I2C < 0)
	{
		MessageBox("CRTC", "I2C初期化エラー");
		return false;
	}
	
	// RTCのリセット
	if (!Reset())
	{
		MessageBox("CRTC", "RTCのリセットに失敗");
		return false;
	}
	// 現在日時を設定する
	if (!SetCurrent())
	{
		MessageBox("CRTC", "RTCの日時設定に失敗");
		return false;
	}

	return true;
}

// ●RTCをリセットする
bool CRTC::Reset(void)
{
	if (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_CTRL, RX8900_BTC_CTRL_CSEL0) < 0)
		return false;
	if (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_EXT, RX8900_BTC_EXT_WADA) < 0)
		return false;
	if (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_FLAG, 0x00) < 0)
		return false;

	return true;
}

// ●現在日時をセットする
bool CRTC::SetCurrent(void)
{
	time_t tCurrent = time(NULL);
	struct tm* pTm = localtime(&tCurrent);

	if (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_CTRL, RX8900_BTC_CTRL_CSEL0 | RX8900_BTC_CTRL_RESET) < 0)
		return false;

	if (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_SEC, bin2bcd(pTm->tm_sec)) < 0)
		return false;
	if (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_MIN, bin2bcd(pTm->tm_min)) < 0)
		return false;
	if (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_HOUR, bin2bcd(pTm->tm_hour)) < 0)
		return false;

	if (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_WEEK, 1 << (pTm->tm_wday)) < 0)
		return false;
	if (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_DAY, bin2bcd(pTm->tm_mday)) < 0)
		return false;
	if (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_MONTH, bin2bcd(pTm->tm_mon + 1)) < 0)
		return false;
	if (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_YEAR, bin2bcd(pTm->tm_year % 100)) < 0)
		return false;

	return true;
}

// ●曜日値を取得する
static int rx8900_get_week_day(BYTE reg_week_day)
{
	int i, tm_wday = -1;
	
	for ( i=0; i < 7; i++ )
	{
		if ( reg_week_day & 1 )
		{
			tm_wday = i;
			break;
		}
		reg_week_day >>= 1;
	}
	
	return 	tm_wday;
}

// ●日時を取得する
void CRTC::GetTime(struct tm* pTm)
{
	pTm->tm_sec  = bcd2bin((BYTE)wiringPiI2CReadReg8(m_I2C, RX8900_BTC_SEC) & 0x7F);
	pTm->tm_min  = bcd2bin((BYTE)wiringPiI2CReadReg8(m_I2C, RX8900_BTC_MIN) & 0x7F);
	pTm->tm_hour = bcd2bin((BYTE)wiringPiI2CReadReg8(m_I2C, RX8900_BTC_HOUR) & 0x3F);
	pTm->tm_wday = rx8900_get_week_day( (BYTE)wiringPiI2CReadReg8(m_I2C, RX8900_BTC_WEEK) & 0x7F);
	pTm->tm_mday = bcd2bin((BYTE)wiringPiI2CReadReg8(m_I2C, RX8900_BTC_DAY) & 0x3F);
	pTm->tm_mon  = bcd2bin((BYTE)wiringPiI2CReadReg8(m_I2C, RX8900_BTC_MONTH) & 0x1F) - 1;
	pTm->tm_year = bcd2bin((BYTE)wiringPiI2CReadReg8(m_I2C, RX8900_BTC_YEAR));
	
	if (pTm->tm_year < 70)
		pTm->tm_year += 100;
}

// ●アラームをセットする
void CRTC::SetTimer(time_t tTimer)
{
	if (m_tTimer == tTimer)
		return;

	ClearTimer();

	bool bErr = false;

	// 日時を設定する
	struct tm tmAlarm = *localtime(&tTimer);
	bErr |= (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_ALARM_MIN, bin2bcd(tmAlarm.tm_min)) < 0);
	bErr |= (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_ALARM_HOUR, bin2bcd(tmAlarm.tm_hour)) < 0);
	bErr |= (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_ALARM_WEEK_OR_DAY, bin2bcd(tmAlarm.tm_mday)) < 0);

	// INTピンを有効にする
	BYTE bCtrl = (BYTE)wiringPiI2CReadReg8(m_I2C, RX8900_BTC_CTRL);
	bCtrl |= RX8900_BTC_CTRL_AIE;
	bErr |= (wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_CTRL, bCtrl) < 0);

	if (bErr)
	{
		MessageBox("CRTC", "アラームの設定に失敗");
		return;
	}

	m_tTimer = tTimer;

	// ログ記録
	int nLast = g_SaveData.ReadInt("LastLog", 0);
	if (++nLast >= 10)
		nLast = 1;

	time_t tCurrent = time(NULL);
	struct tm tmCur = *localtime(&tCurrent);

	char log[64];
	snprintf(log, sizeof(log), "%02d/%02d %02d:%02d:%02d -> %02d/%02d %02d:%02d:00",
		tmCur.tm_mon + 1, tmCur.tm_mday, tmCur.tm_hour, tmCur.tm_min, tmCur.tm_sec, 
			tmAlarm.tm_mon + 1, tmAlarm.tm_mday, tmAlarm.tm_hour, tmAlarm.tm_min);

	char key[32];
	snprintf(key, sizeof(key), "TimerLog%d", nLast);
	
	g_SaveData.WriteStr(key, log);
	g_SaveData.WriteInt("LastLog", nLast);
}

// ●アラームをクリアする
void CRTC::ClearTimer(void)
{
	// AIEビットのクリア
	BYTE val = (BYTE)wiringPiI2CReadReg8(m_I2C, RX8900_BTC_CTRL);
	val &= ~RX8900_BTC_CTRL_AIE;
	wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_CTRL, val);

	// AFビットのクリア
	val = (BYTE)wiringPiI2CReadReg8(m_I2C, RX8900_BTC_FLAG);
	val &= ~RX8900_BTC_FLAG_AF;
	wiringPiI2CWriteReg8(m_I2C, RX8900_BTC_FLAG, val);

	m_tTimer = 0;
}

// ●温度を取得する
double CRTC::GetTempture(void)
{
	double TEMP = wiringPiI2CReadReg8(m_I2C, RX8900_EXT_TEMP);

	return (TEMP * 2 - 187.19) / 3.218;
}
