/****************************************************************************

 ■PhotoFrame
	スケジューラ
        Copyright(C) 2018 Denshi Usagi

****************************************************************************/

#include "Scaduler.h"
#include "main.h"
#include "SDL2/SDL.h"
#include <stdio.h>
#include <limits.h>
#include <assert.h>

#define START_MARGIN	5

//---------------------------------------------------------------------------

// ●コンストラクタ
CScaduler::CScaduler()
{
	m_Current = -1;
}

// ●デストラクタ
CScaduler::~CScaduler()
{
	for(scd_vect::iterator it = m_List.begin(); it != m_List.end(); ++it)
		delete *it;
}

// ●初期化
bool CScaduler::Initialize()
{
	// RTC初期化
	if (!m_RTC.Initialize())
		return false;

	m_RTC.ClearTimer();
	
	// スケジュール設定のロード
	Load();
	
	return true;
}

// ●スケジュール
EScadule CScaduler::Scadule(String* pInfo)
{
	pInfo->clear();

	// 次の候補を探す(次回起動のアラームもセット)
	int nOccur = -1;
	SScadule* pNext;

	time_t tCurrent = time(NULL);
	int nIndex = Next(tCurrent);

	if (nIndex != -1)
	{
		pNext = m_List[nIndex];
		int tDiff = pNext->tNext - tCurrent;
		if (tDiff <= START_MARGIN)
		{
			// 今すぐ実行すべき場合(サウンド再生)
			if (pNext->eType == SC_PlaySound)
			{
				*pInfo = pNext->sInfo;
				pNext->tLast = pNext->tNext;
				if (pNext->tLast < tCurrent)
					pNext->tLast = tCurrent;
				return SC_PlaySound;
			}
			nOccur = nIndex;
		}
		else if (tDiff < (5 * 60))
		{
			// 実行まで5分未満なら終了せずに待つ
			nOccur = -2;
		}
	}

	if (m_Current != -1)
	{
		// 現在実行中のスケジュールがある場合
		SScadule* pCurrent = m_List[m_Current];
		*pInfo = pCurrent->sInfo;
		if (pNext->tLast < tCurrent)
			pNext->tLast = tCurrent;
		return pCurrent->eType;
	}
	if (nOccur >= 0)
	{
		// 今すぐ実行すべき場合
		if (pNext->eType != SC_PotoDisplay)
		{
			// 特別写真の場合
			*pInfo = pNext->sInfo;
			pNext->tLast = pNext->tNext;
			if (pNext->tLast < tCurrent)
				pNext->tLast = tCurrent;
			m_Current = nOccur;
		}
		return pNext->eType;
	}
	if (nOccur == -2) {
		return SC_Wait;
	}

	// 終了(スケジュールなし)
	return SC_Ending;
}

// ●スケジュール実行完了通知
void CScaduler::Complete(void)
{
	m_Current = -1;
}

// ●スケジュールをロードする
void CScaduler::Load(void)
{
	// 通常動作
	LoadPhoto();

	time_t tCurrent = time(NULL);

	// 特別動作(特別写真)
	LoadSpecial("Special", SC_SpecialPhoto, tCurrent);

	// 特別動作(サウンド再生)
	LoadSpecial("Sound", SC_PlaySound, tCurrent);
}

// ●スケジュールをロードする(通常動作)
void CScaduler::LoadPhoto(void)
{
	SScadule* pScadule = new SScadule;
	pScadule->eType = SC_PotoDisplay;
	pScadule->tLast = 0;
	pScadule->tNext = 0;

	// sTimeにフラグを結合して格納する
	for(int i = 0; i <= 23; i++)
	{
		char buf[32];
		snprintf(buf, sizeof(buf), "Photo%02d", i);

		String flag = g_Settings.ReadStr(buf, "");
		if (flag != "0" && flag != "1")
		{
			MessageBox("CScaduler", "設定に誤りがある");
			flag = "1";
		}
		pScadule->sTime += flag;
	}

	m_List.push_back(pScadule);
}

// ●スケジュールをロードする(特別動作)
void CScaduler::LoadSpecial(PCSTR pNameBase, EScadule eType, time_t tCurrent)
{
	for(int i = 0; i < 100; i++)
	{
		char buf[32];
		snprintf(buf, sizeof(buf), "%s%d", pNameBase, i);

		String entry = g_Settings.ReadStr(buf, "");
		if (entry != "")
		{
			if (entry.length() < 28
				|| entry[4] != '/' || entry[7] != '/' || entry[10] != ' '
				|| entry[13] != ':' || entry[16] != ':' || entry[19] != ' ')
			{
				MessageBox("CScaduler", "設定に誤りがある");
				continue;
			}

			SScadule* pScadule = new SScadule;
			pScadule->eType = eType;
			pScadule->sName = buf;
			pScadule->sTime = entry.substr(0, 19);
			pScadule->sInfo = entry.substr(20);
			pScadule->tLast = tCurrent + START_MARGIN;
			pScadule->tNext = 0;

			m_List.push_back(pScadule);
		}
	}
}

// ●スケジュール時間を更新して次の候補を取得する
int CScaduler::Next(time_t tCurrent)
{
	// 次回起動時間を計算してアラームをセットしておく
	int nIndex = CalcNext(true, tCurrent);
	if (nIndex != -1)
	{
		// 90秒前に起動
		// ただし、RTC(RX8900)は秒は指定できないので
		// 設定時刻によって実際には90秒～149秒前になる
		m_RTC.SetTimer(m_List[nIndex]->tNext - 90);
	}	
	else {
		m_RTC.ClearTimer();
	}

	// 次に実行すべきスケジュール 
	return CalcNext(false, tCurrent);
}

// ●次の実行日時を計算する
int CScaduler::CalcNext(bool bFuture, time_t tCurrent)
{
	if (m_List.size() == 0)
		return -1;

	// 全てのスケジュールに次の実行日時をセット
	for(scd_vect::iterator it = m_List.begin(); it != m_List.end(); ++it)
	{
		SScadule* pScadule = *it;
		if (pScadule->eType == SC_PotoDisplay)
			NextPhoto(pScadule, bFuture, tCurrent);
		else
			NextSpecial(pScadule);
	}

	// 最も近い実行時間のスケジュールを取得する
	return GetClosestTime();
}

// ●次の実行日時をセット(通常動作)
void CScaduler::NextPhoto(SScadule* pScadule, bool bFuture, time_t tCurrent)
{
	struct tm tmNext = *localtime(&tCurrent);
	if (!bFuture)
	{
		// 現在時が有効に設定されていれば常に実行する
		if (pScadule->sTime[tmNext.tm_hour] == '1')
		{
			pScadule->tNext = tCurrent;
			return;
		}
	}

	// 次の開始時間(立ち上がり)を探す
	for (int i = 0, j = tmNext.tm_hour; i < 24; i++, j++)
	{
		if (pScadule->sTime[j % 24] == '0')
		{
			for (i = 0, j++; i < 23; i++, j++)
			{
				if (pScadule->sTime[j % 24] == '1')
				{
					tmNext.tm_sec = 0;
					tmNext.tm_min = 0;
					tmNext.tm_hour = j % 24;

					time_t tNext = mktime(&tmNext);
					if (tNext < tCurrent)
					{
						++tmNext.tm_mday;
						tNext = mktime(&tmNext);
						assert(tNext >= tCurrent);
					}
					pScadule->tNext = tNext;
					return;
				}
			}
			// 全て0(実行することはない)
			pScadule->tNext = 0;
			return;
		}
	}

	// 全て1(常に実行する)
	pScadule->tNext = tCurrent;
}

// ●次の実行日時をセット(特別動作)
void CScaduler::NextSpecial(SScadule* pScadule)
{
	time_t tStart = pScadule->tLast + 1;
	struct tm tmNext = *localtime(&tStart);

	struct tm tmAdv;
	memset(&tmAdv, 0, sizeof(tmAdv));

	PCSTR p = pScadule->sTime.c_str();

	#define ISAST(x) (p[18 - x] == '*')
	#define TONUM(x) (p[18 - x] - '0')
	
	bool bAdv = true;
	for(;;)
	{
		// 秒
		if (ISAST(0)) { tmAdv.tm_sec = 1; break; }
		tmNext.tm_sec = (tmNext.tm_sec / 10) * 10 + TONUM(0);
		if (ISAST(1)) { tmAdv.tm_sec = 10; break; }
		tmNext.tm_sec = TONUM(1) * 10 + (tmNext.tm_sec % 10);

		// 分
		if (ISAST(3)) { tmAdv.tm_min = 1; break; }
		tmNext.tm_min = (tmNext.tm_min / 10) * 10 + TONUM(3);
		if (ISAST(4)) { tmAdv.tm_min = 10; break; }
		tmNext.tm_min = TONUM(4) * 10 + (tmNext.tm_min % 10);

		// 時
		if (ISAST(6)) { tmAdv.tm_hour = 1; break; }
		tmNext.tm_hour = (tmNext.tm_hour / 10) * 10 + TONUM(6);
		if (ISAST(7)) { tmAdv.tm_hour = 10; break; }
		tmNext.tm_hour = TONUM(7) * 10 + (tmNext.tm_hour % 10);

		// 日
		if (ISAST( 9)) { tmAdv.tm_mday = 1; break; }
		tmNext.tm_mday = (tmNext.tm_mday / 10) * 10 + TONUM( 9);
		if (ISAST(10)) { tmAdv.tm_mday = 10; break; }
		tmNext.tm_mday = TONUM(10) * 10 + (tmNext.tm_mday % 10);

		// 月
		if (ISAST(12)) { tmAdv.tm_mon = 1; break; }
		tmNext.tm_mon = ((tmNext.tm_mon + 1) / 10) * 10 + TONUM(12) - 1; 
		if (ISAST(13)) { tmAdv.tm_mon = 10; break; }
		tmNext.tm_mon = TONUM(13) * 10 + ((tmNext.tm_mon + 1) % 10) - 1;

		// 年(下2桁のみ有効)
		if (ISAST(15)) { tmAdv.tm_year = 1; break; }
		tmNext.tm_year = ((tmNext.tm_year + 1900) / 10) * 10 + TONUM(15) - 1900;

		if (ISAST(16)) { tmAdv.tm_year = 10; break; }
		tmNext.tm_year = ((tmNext.tm_year + 1900) / 100) * 100
			+ TONUM(16) * 10 + ((tmNext.tm_year + 1900) % 10) - 1900;
		
		bAdv = false;
		break;
	}

	time_t tNext = mktime(&tmNext);
	if (tNext < tStart)
	{
		if (bAdv)
		{
			// mktimeの、各フィールドが範囲外でも正規化してくれる機能を利用する
			tmNext.tm_sec += tmAdv.tm_sec;
			tmNext.tm_min += tmAdv.tm_min;
			tmNext.tm_hour += tmAdv.tm_hour;
			tmNext.tm_mday += tmAdv.tm_mday;
			tmNext.tm_mon += tmAdv.tm_mon;
			tmNext.tm_year += tmAdv.tm_year;
			tNext = mktime(&tmNext);
			assert(tNext >= tStart);
		}
		else
		{
			// 実行予定なし
			tNext = 0;
		}
	}

	pScadule->tNext = tNext;
}

// ●最も近い実行時間のスケジュールを取得する
int CScaduler::GetClosestTime(void)
{
	time_t tMin = (time_t)INT_MAX;
	int nIndex = -1;

	for(int i = 0, j = m_List.size(); i < j; i++)
	{
		SScadule* pScadule = m_List[i];
		if (pScadule->tNext == 0)
			continue;

		if (pScadule->tNext < tMin)
		{
			tMin = pScadule->tNext;
			nIndex = i;
		}
		else if (pScadule->tNext == tMin)
		{
			// 時間がかぶった場合
			// EScadule値が大きいほど優先順位が高い
			// 同じEScadule値なら番号が小さい(先に登録した)ものが優先順位が高い
			if (pScadule->eType > m_List[nIndex]->eType)
				nIndex = i;
		}
	}

	return nIndex;
}
